@@ -34,8 +34,10 @@ impl<T: Send> OneSender<T> {
 }
 impl<T: Send> Drop for OneSender<T> {
     fn drop(&mut self) {
+        let mut inner_guard = self.inner.lock().unwrap();
         self.std_sender.take();
-        let opt_waker = self.inner.lock().unwrap().receiver_waker.take();
+        let opt_waker = inner_guard.receiver_waker.take();
+        drop(inner_guard);
         if let Some(waker) = opt_waker {
             waker.wake();
         }
@@ -57,15 +59,15 @@ impl<T: Send> Future for SendFut<T> {
 
     fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
         // The enclosing Rust-generated `async_send future future prevents re-polling.
-        // It panics at '`async fn` resumed after completion'.
+        // It panics at '`async fn` resumed after completion'.  So this can never panic:
         let value = self.value.take().take().unwrap();
+        let mut inner_guard = self.inner.lock().unwrap();
         match self.std_sender.try_send(value) {
             Ok(()) => Poll::Ready(Ok(())),
             Err(TrySendError::Disconnected(value)) => Poll::Ready(Err(SendError(value))),
             Err(TrySendError::Full(value)) => {
                 self.value.set(Some(value));
-                let waker = cx.waker().clone();
-                self.inner.lock().unwrap().sender_wakers.push(waker);
+                inner_guard.sender_wakers.push(cx.waker().clone());
                 Poll::Pending
             }
         }
@@ -148,11 +150,16 @@ impl<T: Send> SyncSender<T> {
 }
 impl<T: Send> Drop for SyncSender<T> {
     fn drop(&mut self) {
+        let mut inner_guard = self.inner.lock().unwrap();
         self.std_sender.take();
         if Arc::strong_count(&self.inner) < 3 {
             // Either the receiver is already dropped or we are the last sender.
             // Either way, it's safe to wake any receiver.
-            self.wake_receiver();
+            let opt_waker = inner_guard.receiver_waker.take();
+            drop(inner_guard);
+            if let Some(waker) = opt_waker {
+                waker.wake();
+            }
         }
     }
 }
@@ -300,10 +307,15 @@ impl<T: Send> Receiver<T> {
 }
 impl<T: Send> Drop for Receiver<T> {
     fn drop(&mut self) {
-        let waker = self.inner.lock().unwrap().receiver_waker.take();
-        drop(waker);
+        let mut inner_guard = self.inner.lock().unwrap();
         self.std_receiver.take();
-        self.wake_senders();
+        let receiver_waker = inner_guard.receiver_waker.take();
+        let sender_wakers: Vec<Waker> = std::mem::take(&mut inner_guard.sender_wakers);
+        drop(inner_guard);
+        drop(receiver_waker);
+        for waker in sender_wakers {
+            waker.wake();
+        }
     }
 }
 #[doc(hidden)]
@@ -311,13 +323,21 @@ impl<T: Send> Future for Receiver<T> {
     type Output = Result<T, std::sync::mpsc::RecvError>;
 
     fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
+        let mut inner_guard = self.inner.lock().unwrap();
         match self.std_receiver.as_ref().unwrap().try_recv() {
             Ok(value) => Poll::Ready(Ok(value)),
             Err(TryRecvError::Disconnected) => Poll::Ready(Err(RecvError)),
             Err(TryRecvError::Empty) => {
                 let waker = cx.waker().clone();
-                let _opt_waker = self.inner.lock().unwrap().receiver_waker.replace(waker);
-                Poll::Pending
+                if Arc::strong_count(&self.inner) < 2 {
+                    // Last sender dropped.
+                    Poll::Ready(Err(RecvError))
+                } else {
+                    let opt_waker = inner_guard.receiver_waker.replace(waker);
+                    drop(inner_guard);
+                    drop(opt_waker);
+                    Poll::Pending
+                }
             }
         }
     }
